module Builtins where

import AST

data BuiltinId = UtilsPrint | UtilsPrintInt | StringIntValue
               deriving (Eq,Show)

builtinId :: ClassName -> MethodName -> Maybe BuiltinId
builtinId c m =
    case (c,m) of
      ("Utils", "print") -> Just UtilsPrint
      ("Utils", "printInt") -> Just UtilsPrintInt
      ("String", "intValue") -> Just StringIntValue
      _ -> Nothing
          
-- NOTE: if you change the builtin program code, you also have to
-- updated "callBuiltinMethods" in the Execution module.
builtinProgCode :: String
builtinProgCode = 
    "class Object {}\n" ++ 
    "class Utils {\n" ++
    "  void print(Object o) {}\n" ++
    "  void printInt(int i) {}\n" ++
    "}\n" ++
    "class String {\n" ++
    "  int intValue() { return 0; }\n" ++
    "  String concat(String other) { return this + other; }\n" ++
    "}\n" ++
    "class Integer {\n" ++
    "  int __i;\n" ++
    "  Integer(int i) { this.__i = i; }\n" ++ 
    "  int intValue() { return __i; }\n" ++
    "}"

builtinJavaCode :: String
builtinJavaCode = 
    "class GIUtils {\n" ++ 
    "  void print(Object o) { System.out.println(o); }\n" ++
    "  void printInt(int i)   { System.out.println(i); }\n" ++
    "}\n" ++
    "class GIString {\n" ++
    "  String __s;\n" ++
    "  GIString(String s) { __s = s; }\n" ++
    "  public String toString()  { return __s; }\n" ++
    "  GIString concat(GIString other) { return new GIString(__s + other.toString()); }\n" ++
    "  int intValue()     { return Integer.valueOf(__s); }\n" ++
    "}\n" ++
    "class GIInteger {\n" ++
    "  int __i;\n" ++
    "  GIInteger(int i)    { this.__i = i; }\n" ++ 
    "  int intValue()    { return __i; }\n" ++
    "  public String toString() { return \"Integer(__i=\"+__i+\")\"; }\n" ++
    "}"

fixBuiltinProg :: Program -> Program
fixBuiltinProg (Program defs) = Program (map fix defs)
    where
      fix (ClassDef cls) | class_name cls == "Object"
          = ClassDef (cls { class_super = Nothing })
      fix d = d
    
isBuiltinClass :: ClassName -> Bool
isBuiltinClass x = x `elem` ["Object", "Utils", "String", "Integer"]