module Dispatch where

import Control.Monad
import Data.Set (Set)
import qualified Data.Set as Set
import Data.List

import GIMonad
import AST
import Utils
import Types

disp :: IfaceName -> Int -> GI Bool
disp ifaceName j =
    do (Interface _ _ ys rs _ _ rcsigs) <- findIface ifaceName
       let b1 = all (\k -> if k == j then True
                              else dispRcsig (ys !! (j-1)) 
                                     (findRcsig rcsigs (ys !! (k-1))))
                    [1..(length ys)]
       b2 <- allM (dispConstr (ys !! (j-1))) rs
       return (b1 && b2)
    where
      findRcsig rcsigs y = find (\sig -> rcsig_name sig == y) rcsigs
      dispRcsig _ Nothing = True
      dispRcsig y (Just sig) = all (dispMSig y) (rcsig_sigs sig)
      dispMSig y msig = y `freeIn` (map argType (ms_args msig))
      dispConstr y (Impl gs (TyIface ifaceName' _)) =
          do allM (\ k -> case gs !! (k-1) of
                            TyVar x | x == y -> disp ifaceName' k
                            _ -> return True)
                  [1..(length gs)]
      
dispSet :: IfaceName -> GI (Set Int)
dispSet i =
    do iface <- findIface i
       let n = length (iface_impl_types iface)
       foldM (\set j -> do b <- disp i j
                           if b then return (Set.insert j set) else return set)
             Set.empty [1..n]

dispList :: IfaceName -> GI [Int]
dispList i = 
    do set <- dispSet i
       return (Set.toList set)

ndispList :: IfaceName -> GI [Int]
ndispList i =
    do iface <- findIface i
       let n = length (iface_impl_types iface)
       set <- dispSet i
       return [ k | k <- [1..n], not (k `Set.member` set) ]
