module Inheritance where

import Data.Maybe
import Data.List 

import AST
import GIMonad
import Types
import Utils

allSuperClasses :: ClassName -> GI [ClassName]
allSuperClasses name =
    do c <- findClass name
       case class_super c of
         Nothing -> return []
         Just (TyClass name' _) ->
             do supers <- allSuperClasses name'
                return (name' : supers)

upOneClass :: TyN -> GI (Maybe TyN)
upOneClass n =
    do c <- findClass (tyc_name n)
       case class_super c of
         Nothing -> return Nothing
         Just super ->
             let xs = class_tyargs c
                 ts = tyc_args n
             in return (Just $ apply (mkSubst xs ts) super)

allSuperInterfaces :: TyK -> GI [TyK]
allSuperInterfaces k =
    do l <- upOneIface k
       ls <- mapM allSuperInterfaces l
       return (nub (l ++ concat ls))

upOneIface :: TyK -> GI [TyK]
upOneIface k =
    do i <- findIface (tyi_name k)
       case iface_impl_types i of
         [y] ->
             let xs = iface_tyargs i
                 ts = tyi_args k
                 subst = mkSubst xs ts
             in return (mapMaybe (f y subst) (iface_impl_cons i))
         _ -> return []
    where f y subst (Impl [TyVar z] k) | y == z = Just (apply subst k)
          f _ _ _                               = Nothing

extendsClass :: TyN -> TyN -> GI Bool
extendsClass n1 n2 =
    if n1 == n2 then return True
       else do mn1' <- upOneClass n1
               case mn1' of
                 Nothing -> return False
                 Just n1' -> n1' `extendsClass` n2

extendsIface :: TyK -> TyK -> GI Bool
extendsIface k1 k2 =
    if k1 == k2 then return True
       else do k1's <- upOneIface k1
               anyM (\k -> k `extendsIface` k2) k1's

extendsIfaceName :: IfaceName -> IfaceName -> GI Bool
extendsIfaceName i j =
    if i == j then return True
       else do i's <- up i
               anyM (\i' -> i' `extendsIfaceName` j) i's
    where up i = 
              do iface <- findIface i
                 case iface_impl_types iface of
                   [_] -> return $ map getIfaceName (iface_impl_cons iface)
                   _ -> return []
          getIfaceName (Impl _ (TyIface name _)) = name