module MethodTypes where

import Control.Monad
import qualified Data.Set as Set
import Data.List
import Data.Maybe

import AST
import GIMonad
import Inheritance
import Environment
import Utils
import Types
import Subtyping
import Logging
import Pretty

mtypeClass :: MethodName -> TyN -> GI MethodSigG
mtypeClass m n'@(TyClass c' _) = loop n'
    where
      loop n@(TyClass c ts) =
          do mmdef <- findMDef c m
             case mmdef of
               Just mdef -> 
                   do cls <- findClass c
                      let xs = class_tyargs cls
                      return (apply (mkSubst xs ts) (md_sig mdef))
               Nothing ->
                   do mn' <- upOneClass n
                      case mn' of
                        Just n' -> mtypeClass m n'
                        Nothing -> failGI ("Type error: method " ++ m ++ 
                                           " not found in class " ++ c')

mtypeIface :: TEnv -> MethodName -> Type -> [Type] -> GI (MethodSigG, IfaceName)
mtypeIface tenv m t ts =
    do debug ("mtypeIface, m = " ++ show m ++ ", t = " ++ showPPQ t ++ 
              ", ts = " ++ showPPQ ts)
       res <- findMSig m
       case res of
         Nothing -> 
             do debug ("mtypeIface, interface method " ++ m ++ " not found")
                failGI ("Type error: interface method " ++ m ++ " not found")
         Just (iface,msig,j) ->
             do debug ("mtypeIface, found method " ++ m ++ " in interface " ++ 
                       iface_name iface ++ " at dispatch position " ++ show j)
                debug ("mtypeIface, msig = " ++ showPPQ msig)
                let zs' = iface_tyargs iface
                    zs = iface_impl_types iface
                    l = length zs
                    us = map (fst . unMethodArg) (ms_args msig)
                    u = ms_return msig
                    zj = zs !! (j-1)
                    pQuest = 
                        case u of
                          Just (TyG (TyVar x)) ->
                              case elemIndex x zs of
                                Just i -> Just (i+1)
                                Nothing -> Nothing
                          _ -> Nothing
                    -- \Set{\MayBeNil{V_i}}
                    setVsQuest i = 
                        let usts = if i /= j then (us, ts)
                                      else (TyG (TyVar zj):us, t:ts)
                                   in contrib tenv (zs !! (i-1)) usts
                    -- all candidates for \MayBeNil{V_i}
                    viQuestCand :: Int -> GI [Maybe Type]
                    viQuestCand i = 
                        do res <- setVsQuest i
                           case res of
                             Nothing -> return [Nothing]
                             Just setV -> 
                                 do superTypes <- mapM (allSuperTypes tenv) setV
                                    return (map Just (nub (concat superTypes)))
                -- all candidates for \Multi{\MayBeNil{V}}
                -- vsQuest :: [[Maybe Type]]
                vsQuest <- do let f 0 = return [[]]
                                  f i = do l1 <- f (i-1)
                                           l2 <- viQuestCand i
                                           return [ vi : v1Tovi 
                                                  | v1Tovi <- l1, vi <- l2 ]
                              rev <- f l
                              return $ map reverse rev
                mapM_ (\i -> do x <- setVsQuest i
                                debug ("mtypeIface, setVsQuest " ++ show i ++ 
                                       " = " ++ showPP x)) [1..l]
                mapM_ (\i -> do x <- viQuestCand i
                                debug ("mtypeIface, viQuestCand " ++ show i ++ 
                                       " = " ++ showPP x)) [1..l]
                debug ("mtypeIface, pQuest = " ++ show pQuest)
                debug ("mtypeIface, vsQuest = " ++ showPP vsQuest)
                -- argument of pick-constr
                setM <- do l <- sequence
                                [ entailsNil tenv 
                                  (ImplNil vs (iface_name iface)
                                   (replicate (length (iface_tyargs iface))
                                              Nothing))
                                 | vs <- vsQuest ]
                           return (catMaybes l)
                debug ("mtypeIface, setM = " ++ showPPList setM)
                if null setM
                   then failGI ("Cannot call method " ++ m ++ " of interface "
                                ++ iface_name iface ++ " at dispatch position "
                                ++ show j ++ " on receiver " ++ showPPQ t ++
                                " with arguments " ++ showPPQ ts)
                   else do ImplG ws (TyIface _ ws') <- pickConstr tenv pQuest 
                                                         setM
                           return $ (apply (mkSubst (zs++zs') (ws++ws')) msig,
                                     iface_name iface)
             
mtype :: TEnv -> MethodName -> Type -> [Type] 
      -> GI (MethodSigG, 
             Maybe (Type, [Type], IfaceName))
mtype _ m TyBot _ = failGI ("Type error: receiver of call to method " ++ m
                            ++ " has type BOT")
mtype tenv m t ts =
    chooseGIs [do mn <- bound tenv t
                  case mn of
                    Just n -> do sig <- mtypeClass m n
                                 return (sig, Nothing)
                    _ -> failGI ("Type error: no bound for " ++ showPP t)
              ,do (sig, i) <- mtypeIface tenv m t ts
                  -- check if it's a retro call or not
                  case t of
                    TyK (TyIface j _) ->
                        do b <- extendsIfaceName j i 
                           return $ if b then (sig, Nothing) 
                                         else (sig, Just (t, ts, i))
                    _ -> return (sig, Just (t, ts, i))
              ,failGI ("Type error: no method " ++ show m ++ " found for " ++
                       "receiver type " ++ showPPQ t ++ " and argument types "
                       ++ showPPQ ts)]

bound :: TEnv -> Type -> GI (Maybe TyN)
bound tenv t =
    case t of
      TyBot -> bug ("MethodTypes.bound: TyBot")
      TyG (TyVar x) -> 
          let nbounds = mapMaybe (\t -> case t of
                                          TyG (TyN n) -> Just n
                                          _ -> Nothing) (upperBounds x tenv)
          in if null nbounds then return $ Just tyObjectN
                else do TyN n <- gminimal tenv (map TyN nbounds)
                        return $ Just n
      TyG (TyN n) -> return $ Just n
      TyK _ -> return $ Just tyObjectN
      _ -> return Nothing

contrib :: TEnv -> TyvarName -> ([Type], [Type]) -> GI (Maybe [Type])
contrib tenv x (us, ts) =
    do debug ("contrib, x = " ++ showPP x ++ ", us = " ++ showPPList us ++ 
              ", ts = " ++ showPPList ts)
       let n = length ts
           setT = [ts!!i | i <- [0 .. (n-1)], us!!i == TyG (TyVar x)]
       debug ("contrib, setT = " ++ showPPList setT)
       case setT of
         [] -> return Nothing
         _ -> case elemIndex TyBot setT of
                Just i -> failGI ("Type error: argument number " ++ show (i+1)
                                  ++ " is a dispatch argument but has type BOT")
                _ -> do l <- mub tenv setT
                        return $ Just l

mub :: TEnv -> [Type] -> GI [Type]
mub tenv ts =
    do debug ("mub, ts = " ++ showPPList ts)
       supers <- mapM (allSuperTypes' tenv) ts
       let setV = foldl1 Set.intersection supers
           listV = Set.toList setV
       debug ("mub, listV = " ++ showPPList listV)
       res <- findMins [] listV []
       debug ("mub, res = " ++ showPPList res)
       return res
    where
      findMins _ [] acc = return acc
      findMins before (t:ts) acc =
          do b1 <- allM (notSubOf t) before
             b2 <- allM (notSubOf t) ts
             findMins (t:before) ts (if b1 && b2 then t:acc else acc)
      notSubOf t u = do b <- sub' tenv u t
                        return (not b)

pickConstr :: TEnv -> Maybe Int -> [ConGR] -> GI ConGR
pickConstr _ _ [] = bug ("MethodTypes.pickConstr: argument list empty")
pickConstr _ Nothing (x:_) = return x
pickConstr tenv (Just i) l = findSmallest l 
    where
      us :: [Type]
      us = map (\ (ImplG ts _) -> (ts !! (i-1))) l
      findSmallest :: [ConGR] -> GI ConGR
      findSmallest [] = bug ("MethodTypes.pickConstr: no smallest element")
      findSmallest (c@(ImplG ts _) : cs) =
          do b <- allM (\u -> sub' tenv (ts !! (i-1)) u) us
             if b then return c else findSmallest cs
          
smtype :: TEnv -> MethodName -> TyK -> [Type] -> GI MethodSigG
smtype tenv m k@(TyIface i us) ts =
    do mmsig <- findStaticMSig i m
       case mmsig of
         Nothing -> failGI ("Type error: static method " ++ m ++ 
                            " not defined for interface " ++ i)
         Just msig ->
             do let con = ConGR (ImplG ts k)
                b <- entails tenv con
                unless b $ failGI("Type error: cannot invoke static interface "
                                  ++ "method "
                                  ++ m ++ " on " ++ showPP k ++ 
                                  "[" ++ showPP ts ++ 
                                  "] because constraint " ++ showPP con ++ 
                                  " does not hold")
                iface <- findIface i
                let xs = iface_tyargs iface
                    ys = iface_impl_types iface
                    subst = mkSubst (xs++ys) (us++ts)
                return (apply subst msig)