module NameEnv where

import Prelude hiding (lookup)
import Data.Map (Map)
import qualified Data.Map as Map

import Logging
import Utils

type T k v = [Map k v]

empty :: T k v
empty = []

push :: (Show k, Ord k) => [k] -> [v] -> T k v -> T k v
push ks vs env = if not (uniqueElems ks) 
                    then fatal ("Parse error: variables " ++ show ks ++ 
                                " must be disjoint")
                    else let m = Map.fromList (zip ks vs)
                             in m:env

insert :: (Show k, Ord k) => k -> v -> T k v -> T k v
insert k v env =
    case lookup k env of
      Just _ -> fatal ("Parse error: variable " ++ show k ++ 
                       "already defined")
      Nothing ->
          if null env 
             then bug ("NameEnv.add: no environment present")
             else Map.insert k v (head env) : tail env

pop :: T k v -> T k v
pop [] = bug ("NameEnv.pop: no environment present")
pop (_:xs) = xs

lookup :: Ord k => k -> T k v -> Maybe v
lookup _ [] = Nothing
lookup k (x:xs) = 
    case Map.lookup k x of
      Just v -> Just v
      _ -> lookup k xs

elems :: T k v -> [v]
elems = concatMap Map.elems
