{ 

{-# OPTIONS -fno-warn-name-shadowing -fno-warn-missing-signatures -fno-warn-unused-binds -fno-warn-unused-matches -fno-warn-incomplete-patterns    #-}
-- ^ grr. happy needs them all on one line

module Parser 
    ( parseProgram, parseExpr, parseType, parseConstraint, parseConstraintG ) 
where

import Control.Monad.State

import AST
import Lexer
import ParserUtils
import Logging
import Builtins

}


-- Conflicts
%expect 1  -- dangling else


%name parseProgramIntern start
%name parseExprIntern start_expr
%name parseConstraintIntern start_constr
%name parseConstraintGIntern start_constrG
%name parseTypeIntern start_type

%error { parseError }
%tokentype { Token }
%monad { PM }
%token 
    '('         { T _ (SpecialT '(') }
    ')'         { T _ (SpecialT ')') }
    '{'         { T _ (SpecialT '{') }
    '}'         { T _ (SpecialT '}') }
    '['         { T _ (SpecialT '[') }
    ']'         { T _ (SpecialT ']') }
    ';'         { T _ (SpecialT ';') }
    ','         { T _ (SpecialT ',') }
    '.'         { T _ (SpecialT '.') }
    '='         { T _ (SpecialT '=') }

    'class'           { T _ (KeywordT "class")   }
    'interface'       { T _ (KeywordT "interface")   }
    'implementation'  { T _ (KeywordT "implementation")   }
    'implementation<' { T _ (KeywordT "implementation<")   }
    'extends'         { T _ (KeywordT "extends")   }
    'implements'      { T _ (KeywordT "implements")   }
    'mono'            { T _ (KeywordT "mono")   }
    'where'           { T _ (KeywordT "where")   }
    'static'          { T _ (KeywordT "static")   }
    'void'            { T _ (KeywordT "void")   }
    'receiver'        { T _ (KeywordT "receiver")   }
    'return'          { T _ (KeywordT "return")   }
    'new'             { T _ (KeywordT "new")   }
    'if'              { T _ (KeywordT "if")   }
    'else'            { T _ (KeywordT "else")   }
    'while'           { T _ (KeywordT "while")   }
    'true'            { T _ (KeywordT "true")   }
    'false'           { T _ (KeywordT "false")   }
    'null'            { T _ (KeywordT "null")   }
    'int'             { T _ (KeywordT "int")   }
    'bool'            { T _ (KeywordT "bool")   }

    '+'        { T _ (OpT "+")   }
    '-'        { T _ (OpT "-")   }
    '*'        { T _ (OpT "*")   }
    '/'        { T _ (OpT "/")   }
    '%'        { T _ (OpT "%")   }
    '&&'       { T _ (OpT "&&")   }
    '||'       { T _ (OpT "||")   }
    '!'        { T _ (OpT "!")   }
    '<='       { T _ (OpT "<=")   }
    '<'        { T _ (OpT "<")   }
    '>='       { T _ (OpT ">=")   }
    '>'        { T _ (OpT ">")   }
    '=='       { T _ (OpT "==")   }
    '!='       { T _ (OpT "!=")   }      

    INT            { T _ (NumT $$) }
    STRING         { T _ (StringT $$) }
    VAR            { T _ (VarT _) }
    VAR_TYARG      { T _ (VarTyargT _) }
    EOF            { EOF }
 
-- Precedence (earlier means lower precedence)
%left '||'
%left '&&'
%left '==' '!='
%left '<=' '<' '>=' '>'
%left '+' '-'
%left '*' '/' '%'
%right '!' CAST_PREC NEW_PREC UN_MINUS_PREC
%left '.'

%%

start : program EOF         { $1 }

start_expr : expr EOF       { $1 }

start_constr : constr EOF   { $1 }

start_constrG : constrG EOF { $1 }

start_type : type EOF       { $1 }


-- 
-- Programs
--

program :: { Program }
program : def_list_r       { Program (reverse $1) }

def_list_r :: { [Def] }
def_list_r : {- empty -}       { [] }
           | def_list_r def    { $2 : $1 }

def :: { Def }
def : impl_def     { ImplDef $1 }
    | iface_def    { IfaceDef $1 }
    | class_def    { ClassDef $1 }

-- 
-- Implementations
--

impl_def :: { Implementation }
impl_def : impl_def_header tynk '[' tynk_list_comma_ne ']' constr_clause 
             '{' static_mdefs receiver_defs '}'
           {% undefineTypeParams "implementation" >>
              freshImplId >>= \id ->
              return (Implementation $1 (asIfaceType $2) (map asClassType $4)
                                     $6 $8 $9 id) }
impl_def_header :: { [TyvarName] }
impl_def_header : 'implementation'                     
                    {% debug ("Parsing implementation definition at " 
                             ++ show (posn $1)) >>
                       defineTypeParams "implementation" [] }
                | 'implementation<' tyvar_list_ne_comma '>'  
                    {% debug ("Parsing implementation definition at " 
                             ++ show (posn $1)) >>
                       defineTypeParams "implementation" $2 }

receiver_defs :: { [ReceiverDef] }
receiver_defs : receiver_defs_r         { reverse $1 }
              | mdefs_ne                 { [ReceiverDef $1] }

receiver_defs_r : {- empty -}                    { [] }
                | receiver_defs_r receiver_def   { $2 : $1 }

receiver_def :: { ReceiverDef }
receiver_def : 'receiver' '{' mdefs '}'          { ReceiverDef $3 }

--
-- Interfaces
--

iface_def :: { Interface }
iface_def : 'interface' class_name_and_args iface_impl_types constr_clause
              '{' static_msigs receiver_sigs '}'
            {% handleInterfaceDef $2 $3 $4 $6 $7 }

iface_impl_types :: { ([TyvarName], [ConR]) }
iface_impl_types : {- empty -}  {% defineTypeParams "interface" ["This"] >>=
                                   \tyvars -> return (tyvars, []) }
                 | '[' iface_impl_vars_ne impl_constr_clause ']'
                   { ($2, $3) }

iface_impl_vars_ne : iface_impl_vars_ne_r   
                       {% defineTypeParams "interface" (reverse $1) }

iface_impl_vars_ne_r : VAR                           { [stringOfVar $1] }
                     | iface_impl_vars_ne_r ',' VAR  { stringOfVar $3 : $1 }

receiver_sigs :: { Maybe [ReceiverSig] }
receiver_sigs : {- empty -}           { Nothing }
              | msigs_ne              {% lookupTHIS >>= 
                                        \x -> return (Just [ReceiverSig x $1 ]) }
              | receiver_sigs_ne { Just $1 }

receiver_sigs_ne : receiver_sigs_ne_r                { reverse $1 }

receiver_sigs_ne_r : receiver_sig                    { [$1] }
                   | receiver_sigs_ne_r receiver_sig { $2 : $1 }

receiver_sig : 'receiver' type_var '{' msigs '}'
               { ReceiverSig $2 $4 }

--
-- Classes
--

class_def :: { Class }
class_def : 'class' class_name_and_args extends_clause constr_clause 
              '{' fmcdefs '}'
            {% handleClassDef $1 $2 $3 $4 $6 }

class_name_and_args :: { (String, [TyvarName]) }
class_name_and_args : VAR                          
                      {% let name = stringOfVar $1 in
                         debug ("Parsing class/interface " ++ show name ++ " at "
                               ++ show (posn $1)) >>
                         defineCurrentClassName name >>
                         defineTypeParams "class/interface" [] >>
                         return (name, []) }
                    | VAR_TYARG tyvar_list_ne_comma '>'  
                      {% let name = stringOfVar $1 in
                         debug ("Parsing class/interface " ++ show name ++ " at "
                               ++ show (posn $1)) >>
                         defineCurrentClassName name >>
                         defineTypeParams "class/interface" $2 >>= 
                         \tyvars -> return (name, tyvars) }

extends_clause :: { TyN }
extends_clause : {- empty -}     { TyClass "Object" [] }
               | 'extends' tynk  { asClassType $2 }

--
-- Methods
--

static_msigs :: { [MethodSig] }
static_msigs : static_msigs_r                        { reverse $1 }

static_msigs_r : {- empty -}                         { [] }
               | static_msigs_r static_msig_semi     { $2 : $1 }

msigs :: { [MethodSig] }
msigs : msigs_r                 { reverse $1 }

msigs_ne :: { [MethodSig] }
msigs_ne : msig_semi msigs      { $1 : $2 }

msigs_r : {- empty -}           { [] }
        | msigs_r msig_semi     { $2 : $1 }


static_msig_semi :: { MethodSig }
static_msig_semi : 'static' msig ';'            { $2 }

msig_semi :: { MethodSig }
msig_semi : msig ';'            { $1 }

msig :: { MethodSig }
msig : type_params return_type VAR expr_params constr_clause
         {% undefineTypeParams "method" >> 
            undefineExprParams "method" >> 
            return (MethodSig $1 $2 (stringOfVar $3) (snd $4) $5) }

static_mdefs :: { [MethodDef] }
static_mdefs : static_mdefs_r                   { reverse $1 }

static_mdefs_r : {- empty -}             { [] }
               | static_mdefs_r static_mdef            { $2 : $1 }

fmcdefs :: { [Either FieldDef (Either Constructor MethodDef)] }
fmcdefs : fmcdefs_r                   { reverse $1 }

fmcdefs_r : {- empty -}             { [] }
          | fmcdefs_r fmcdef          { $2 : $1 }

static_mdef :: { MethodDef }
static_mdef : 'static' mdef       { $2 }

fmcdef :: { Either FieldDef (Either Constructor MethodDef) }
fmcdef : mdef  { Right (Right $1) }
       | type_params return_type expr_params '{' stmt_list '}'  
         {% do { undefineTypeParams "method/constructor/field"
               ; undefineExprParams "method/constructor/field"
               ; curName <- getCurrentClassName
               ; case ($1,$2) of
                   ([], Just (TyParse (TyClassIface name []))) | name == curName
                       -> return (Right (Left (Constructor (fst $3) (snd $3) $5)))
                   _ -> fatal ("Parse error around " ++ show (posn $4)) }
         }
       | type_params return_type VAR ';'  
         {% do { undefineTypeParams "method/constructor/field"
               ; case ($1, $2) of
                   ([], Just t) -> return (Left (FieldDef t (stringOfVar $3)))
                   _ -> fatal ("Parse error around " ++ show (posn $4)) }
         }

mdefs :: { [MethodDef] }
mdefs : mdefs_r                   { reverse $1 }

mdefs_r : {- empty -}             { [] }
         | mdefs_r mdef           { $2 : $1 }

mdefs_ne :: { [MethodDef] }
mdefs_ne : mdefs_ne_r                  { reverse $1 }

mdefs_ne_r : mdef                      { [$1] }
           | mdefs_ne_r mdef           { $2 : $1 }


mdef :: { MethodDef }
mdef : type_params return_type VAR expr_params constr_clause '{' stmt_list '}'     
          {% undefineTypeParams "method/constructor/field" >> 
             undefineExprParams "method/constructor/field" >> 
             return (MethodDef (MethodSig $1 $2 (stringOfVar $3) 
                                          (snd $4) $5) (fst $4) $7) }

type_params :: { [TyvarName] }
type_params : {- empty -}                
                {% defineTypeParams "method/constructor/field" [] }
            | '<' tyvar_list_ne_comma '>'      
                {% defineTypeParams "method/constructor/field" $2 }

return_type :: { Maybe Type }
return_type : 'void'           { Nothing }
            | type             { Just $1 }

expr_params :: { (VarName, [MethodArg]) }
expr_params : '(' type_and_var_list ')'
              {% let l = $2 in
                 defineExprParams "method" ("this" : map snd l) >>= 
                 \ (this:vars) -> return (this,map MethodArg (zip (map fst l) vars)) }

type_and_var_list :: { [(Type, String)] }
type_and_var_list : {- empty -}               { [] }
                  | type_and_var_list_ne_r    { reverse $1 }

type_and_var_list_ne_r : type VAR                  
                         { [($1, stringOfVar $2)] }
                       | type_and_var_list_ne_r ',' type VAR
                         { ($3, stringOfVar $4) : $1 }
--     
-- Statements
--

stmt :: { Stmt }
stmt : ';'                                    { Skip }
     | expr ';'                               { StmtExpr $1 }
     | type VAR '=' expr ';'                  
       {% handleLocalDef $1 (stringOfVar $2) $4 }
     | VAR '=' expr ';'                       
       {% handleAssign (stringOfVar $1) $3 }
     | expr '.' VAR '=' expr ';'              
       {% newRef >>= \r -> return (AssignField $1 (stringOfVar $3) $5 r) }
     | 'if' '(' expr ')' block                { If $3 $5 }
     | 'if' '(' expr ')' block 'else' block   { IfElse $3 $5 $7 }
     | 'while' '(' expr ')' block             { While $3 $5 }
     | 'return' expr ';'                      
       {% newRef >>= \r -> return (Return (Just $2) r) }
     | 'return' ';'                           
       {% newRef >>= \r -> return (Return Nothing r) }

block :: { [Stmt] }
block : stmt                      { [$1] }
      | '{' stmt_list '}'         { $2 }

stmt_list :: { [Stmt] }
stmt_list : stmt_list_r           { reverse $1 }

stmt_list_r : {- empty -}         { [] }
            | stmt_list_r stmt    { $2 : $1 }

--
-- Expressions                      
--

expr :: { Expr }
expr : VAR             {% handleExprVar (stringOfVar $1) }
     | INT             { Lit (LitInt $1) }
     | STRING          { Lit (LitString $1) }
     | 'null'          { Lit LitNull }
     | 'true'          { Lit (LitBool True) }
     | 'false'         { Lit (LitBool False) }
     | expr '+' expr   { PrimOp Add [$1,$3] }
     | expr '-' expr   { PrimOp Sub [$1,$3] }
     | expr '*' expr   { PrimOp Mult [$1,$3] }
     | expr '/' expr   { PrimOp Div [$1,$3] }
     | expr '%' expr   { PrimOp Mod [$1,$3] }
     | expr '==' expr  { PrimOp Equal [$1,$3] }
     | expr '!=' expr  { PrimOp NotEqual [$1,$3] }
     | expr '<=' expr  { PrimOp Leq [$1,$3] }
     | expr '<' expr   { PrimOp Lt [$1,$3] }
     | expr '>=' expr  { PrimOp Geq [$1,$3] }
     | expr '>' expr   { PrimOp Gt [$1,$3] }
     | '!' expr        { PrimOp Not [$2] }
     | expr '&&' expr  { PrimOp And [$1,$3] }
     | expr '||' expr  { PrimOp Or [$1,$3] } 
     | expr '.' VAR    {% newRef >>= \t -> return (FAccess $1 (stringOfVar $3) t) }
     | expr '.' var_tyargs expr_args   
       {% newRef >>= \t -> return (MCall $1 (fst $3) (snd $3) $4 t) }
     | var_tyargs expr_args            
       {% handleMCall (fst $1) (snd $1) $2 }
     | tynk '[' type_list_comma_ne ']' '.' var_tyargs expr_args
       {% newRef >>= \t -> return (SMCall (asIfaceType $1) $3 (fst $6) (snd $6) $7 t) }
     | '-' expr             %prec UN_MINUS_PREC 
       { PrimOp UnMinus [$2] }
     | '(' tynk ')' expr    %prec CAST_PREC     
       { Cast (asClassType $2) $4 }
     | 'new' tynk expr_args %prec NEW_PREC      
       {% newRef >>= \r -> return (New (asClassType $2) $3 r) }

-- 
-- Types and constraints
--

type :: { Type }                  
type : VAR                                 {% handleType $1 [] }
     | VAR_TYARG type_args_no_open         {% handleType $1 $2 }
     | 'int'                               { TyPrim TyInt }
     | 'bool'                              { TyPrim TyBool }
            
type_with_posn :: { (Posn, Type) }                  
type_with_posn : VAR                           {% handleType $1 [] >>=
                                                    \t -> return (posn $1, t) }
               | VAR_TYARG type_args_no_open   {% handleType $1 $2  >>=
                                                    \t -> return (posn $1, t) }
               | 'int'                         { (posn $1, TyPrim TyInt) }
               | 'bool'                        { (posn $1, TyPrim TyBool) }
    
tynk :: { TyNK }       
tynk : type_with_posn  { case $1 of
                           (_, TyParse t) -> t
                           (p,t) -> fatal ("Parse error: expected class or interface "
                                           ++ " type at " ++ show p ++ 
                                           ", found: " ++ show t) }

constr_clause :: { [ConP] }
constr_clause : {- empty -}             { [] }
              | 'where' constr_list_ne  { $2 }

constr_list_ne : constr_list_ne_r       { reverse $1 }

constr_list_ne_r : constr                        { [$1] }
                 | constr_list_ne_r ',' constr   { $3 : $1 }

constr :: { ConP }
constr : impl_left 'implements' tynk    { ConR (Impl $1 (asIfaceType $3)) }
       | type_var 'extends' type        { Ext $1 $3 }
       | type_var 'mono'                { Mono $1 }

impl_constr_clause :: { [ConR] }
impl_constr_clause : {- empty -}                  { [] }
                   | 'where' impl_constr_list_ne  { $2 }

impl_constr_list_ne : impl_constr_list_ne_r       { reverse $1 }

impl_constr_list_ne_r : impl_constr                             { [$1] }
                      | impl_constr_list_ne_r ',' impl_constr   { $3 : $1 }

impl_constr :: { ConR }
impl_constr : impl_left 'implements' tynk    { Impl $1 (asIfaceType $3) }

impl_left :: { [TyG] }
impl_left : type_with_posn                { [asTyG (fst $1) (snd $1)] }
          | '[' type_list_comma_ne ']'    { map (asTyG (posn $1)) $2 }

constrG :: { ConGP }
constrG : impl_leftG 'implements' tynk    { ConGR (ImplG $1 (asIfaceType $3)) }
        | type 'extends' type             { ExtG $1 $3 }
        | type 'mono'                     { MonoG $1 }

impl_leftG :: { [Type] }
impl_leftG : type                          { [$1] }
           | '[' type_list_comma_ne ']'    { $2 }

--
-- Utilities
--

expr_args :: { [Expr] }
expr_args : '(' ')'                                    { [] }
          | '(' expr_list_comma_ne ')'                 { $2 }

expr_list_comma_ne : expr_list_comma_ne_r              { reverse $1 }

expr_list_comma_ne_r : expr                            { [$1] }
                     | expr_list_comma_ne_r ',' expr   { $3 : $1 }

type_args_no_open : type_list_comma_ne '>'             { $1 }

type_list_comma_ne : type_list_comma_ne_r              { reverse $1 }

type_list_comma_ne_r : type                            { [$1] }
                     | type_list_comma_ne_r ',' type   { $3 : $1 }

tyvar_list_ne_comma :: { [String] }
tyvar_list_ne_comma : tyvar_list_ne_comma_r       { reverse $1 }

tyvar_list_ne_comma_r : VAR                       { [stringOfVar $1] }
                      | tyvar_list_ne_comma_r ',' VAR   { stringOfVar $3 : $1 }

var_tyargs :: { (String, [Type]) }
var_tyargs : VAR                                       { (stringOfVar $1, []) }
           | VAR_TYARG type_args_no_open               { (stringOfVar $1, $2) }

type_var :: { TyvarName }
type_var : VAR           {% handleTypeVar $1 }

tynk_list_comma_ne :: { [TyNK] }
tynk_list_comma_ne : tynk_list_comma_ne_r                { reverse $1 }

tynk_list_comma_ne_r : tynk                               { [$1] }
                     | tynk_list_comma_ne_r ',' tynk      { $3 : $1 }

{

parseProgram :: [Token] -> IO Program
parseProgram toks = 
    do (Program defs,_) <- runStateT (parseProgramIntern toks) emptyParseState
       Program defs' <- parseBuiltinProgram
       return (fixAST (Program (defs' ++ defs)))

parseBuiltinProgram :: IO Program
parseBuiltinProgram =
    do let toks = scan "<builtins>" builtinProgCode
       res <- runStateT (parseProgramIntern toks) emptyParseState
       return (fixBuiltinProg $ fixAST (fst res))

parseExpr :: [Token] -> [String] -> [String] -> IO Expr
parseExpr toks vars tyvars =
    do let state = mkParseState ("this":vars) tyvars
       res <- runStateT (parseExprIntern toks) state
       return (fst res)

parseType :: [Token] -> [String] -> Program -> IO Type
parseType = genericParse parseTypeIntern

parseConstraint :: [Token] -> [String] -> Program -> IO ConP
parseConstraint = genericParse parseConstraintIntern

parseConstraintG :: [Token] -> [String] -> Program -> IO ConGP
parseConstraintG = genericParse parseConstraintGIntern

}
