module Positions where

import Data.Maybe

import GIMonad
import AST
import Types
import Utils
import Logging

data PlusMinus = Plus | Minus

posPlus :: IfaceName -> Int -> GI Bool
posPlus iname j = posGeneric iname j Plus

posMinus :: IfaceName -> Int -> GI Bool
posMinus iname j = posGeneric iname j Minus

posGeneric :: IfaceName -> Int -> PlusMinus -> GI Bool
posGeneric iname j pm =
    do iface <- findIface iname
       let ys = iface_impl_types iface
           y = if j < 1 || j > length ys 
                  then fatal ("Positions.posGeneric: invalid index " ++ show j)
                  else ys !! (j-1)
       andM [return $ all (posMSigGeneric y pm) (iface_static_sigs iface)
            ,return $ all (posRCSigGeneric y pm) (iface_receiver_sigs iface)
            ,allM (posConsGeneric y pm) (iface_impl_cons iface)
            ,return $ not (y `freeIn` (iface_cons iface))]
    where
      posMSigGeneric y Plus sig = not (y `freeIn` ms_args sig)
      posMSigGeneric y Minus sig = not (y `freeIn` ms_return sig)
      posRCSigGeneric y pm rcsig = all (posMSigGeneric y pm) (rcsig_sigs rcsig)
      posConsGeneric y pm (Impl gs (TyIface iname _)) =
          allM (\ (g,j) -> 
                    case g of
                      TyVar x | x == y -> posGeneric iname j pm
                      _ -> return True)
               (zip gs [1..])
          
positions :: IfaceName -> PlusMinus -> GI [Int]
positions i pm = 
    do iface <- findIface i
       let n = length (iface_impl_types iface)
       ml <- mapM (\k -> do b <- posGeneric i k pm
                            return $ if b then Just k else Nothing) [1..n]
       return (catMaybes ml)

notPositions :: IfaceName -> PlusMinus -> GI [Int]
notPositions i pm =
    do iface <- findIface i
       let n = length (iface_impl_types iface)
       l <- positions i pm
       return [ k | k <- [1..n], not (k `elem` l)]
