module StatementTyping where

import Control.Monad

import Environment
import AST
import ExpressionTyping
import GIMonad
import Pretty
import Logging

stmtType :: TEnv -> VEnv 
         -> Maybe Type    -- the type expected from a return expression
         -> Stmt
         -> GI (VEnv,     -- a possibly extended varialbe environment
                Bool)     -- does the statement definitely return?
stmtType tenv venv retType stmt = 
    addTrace ("Checking statement " ++ showPPQ stmt ++ " failed")
             (do debug ("Checking statement " ++ showPPQ stmt)
                 res <- stmtType' tenv venv retType stmt
                 debug ("Statement " ++ showPPQ stmt ++ " typechecks")
                 return res)

stmtType' :: TEnv -> VEnv -> Maybe Type -> Stmt -> GI (VEnv, Bool)
stmtType' tenv venv retType stmt = 
    case stmt of
      StmtExpr e ->
          do catchGI ((==) VoidReturn) 
                 (do _ <- exprType tenv venv e
                     return ())
                 (return ())
             return (venv, False)
      DefineLocal t x e r ->
          do t' <- exprType tenv venv e
             [coercion] <- assertSub tenv [t'] [t]
             fillRef r coercion
             let venv' = extendVEnv venv [x] [t]
             return (venv', False)
      AssignLocal x e r ->
          case typeOfVar x venv of
            Nothing -> failGI ("Type error: undefined variable " ++ showPP x)
            Just t ->
                do t' <- exprType tenv venv e
                   [coercion] <- assertSub tenv [t'] [t]
                   fillRef r coercion
                   return (venv, False)
      AssignField e f e' r ->
          do t <- fieldType tenv venv e f
             t' <- exprType tenv venv e'
             [coercion] <- assertSub tenv [t'] [t]
             fillRef r coercion
             return (venv, False)
      If e stmts ->
          do t <- exprType tenv venv e
             unless (t == tyBool) $ failGI ("Type error: condition of if must " 
                                            ++ "have type " ++ showPPQ tyBool ++
                                           " but has type " ++ showPPQ t)
             _ <- stmtsTypes tenv venv retType stmts
             return (venv, False)
      IfElse e stmts stmts' ->
          do t <- exprType tenv venv e
             unless (t == tyBool) $ failGI ("Type error: condition of if must " 
                                            ++ "have type " ++ showPPQ tyBool ++
                                           " but has type " ++ showPPQ t)
             (_, returns) <- stmtsTypes tenv venv retType stmts
             (_, returns') <- stmtsTypes tenv venv retType stmts'
             return (venv, returns && returns')
      While e stmts ->
          do t <- exprType tenv venv e
             unless (t == tyBool) $ failGI ("Type error: condition of while " ++
                                            "must have type " ++ 
                                            showPPQ tyBool ++ " but has type " 
                                            ++ showPPQ t)
             _ <- stmtsTypes tenv venv retType stmts
             return (venv, False)
      Return me r ->
          case (me,retType) of
            (Nothing,Nothing) -> return (venv,True)
            (Just e, Just t) ->
                do t' <- exprType tenv venv e
                   [coercion] <- assertSub tenv [t'] [t]
                   fillRef r coercion
                   return (venv,True)
            (Just _,Nothing) ->
                failGI ("Type error: no return value expected")
            (Nothing,Just _) ->
                failGI ("Type error: return value expected but none provided")
      Skip -> return (venv,False)
      
stmtsTypes :: TEnv -> VEnv -> Maybe Type -> [Stmt] -> GI (VEnv, Bool)
stmtsTypes _ venv _ [] = return (venv, False)
stmtsTypes tenv venv retType (s:ss) =
    do (venv', returns) <- stmtType tenv venv retType s
       (venv'', returns') <- stmtsTypes tenv venv' retType ss
       return (venv'', returns || returns')