module Unification (

  UP, unify, unify'

) where

import AST
import Types

type UP = [(Type,Type)]

unify  :: UP -> Maybe TySubst
unify up = unify' (\_ -> True) up

unify' :: (TyvarName -> Bool)  -- returns True of the given type variable
                               -- is instantiable
       -> UP ->  Maybe TySubst
unify' p up = unify'' p emptySubst up

unify'' :: (TyvarName -> Bool) -> TySubst -> UP ->  Maybe TySubst
unify'' _ subst [] = return subst
unify'' p subst ((t,u) : rest)
    | t == u = unify'' p subst rest  -- DELETE
    | otherwise =
        case (t,u) of
          -- DECOMPOSE
          (TyG (TyN (TyClass name args)),
           TyG (TyN (TyClass name' args'))) 
              | name == name' -> unify'' p subst (zip args args' ++ rest)
          (TyK (TyIface name args),
           TyK (TyIface name' args'))
              | name == name' -> unify'' p subst (zip args args' ++ rest)
          -- special case: two variables (needed to support to extra predicate p)
          (TyG (TyVar x), TyG (TyVar y)) ->
              case () of
              _| p x -> eliminate x u
               | p y -> eliminate y t
               | otherwise -> Nothing
          -- ORIENT
          (t, TyG (TyVar _)) -> unify'' p subst ((u,t) : rest)
          -- ELIMINATE
          (TyG (TyVar x), u) | not (x `freeIn` u) && p x -> eliminate x u
          -- FAILURE
          _ -> Nothing
    where
      eliminate x u =
          let s = mkSubst [x] [u]
          in unify'' p (s `composeSubst` subst) (apply s rest)

