module Utils where

import Control.Monad
import Data.List
import Data.Set (Set)
import qualified Data.Set as Set

uniqueElems :: Eq a => [a] -> Bool
uniqueElems l = nub l == l

anyM :: Monad m => (a -> m Bool) -> [a] -> m Bool
anyM p l = orM (map p l)

allM :: Monad m => (a -> m Bool) -> [a] -> m Bool
allM p l = andM (map p l)

orM :: Monad m => [m Bool] -> m Bool
orM [] = return False
orM (m:ms) = do b <- m
                if b then return True else orM ms
           

andM :: Monad m => [m Bool] -> m Bool
andM [] = return True
andM (m:ms) = do b <- m
                 if not b then return False else andM ms

quote :: String -> String
quote s = "``" ++ s ++ "''"

findFirstJust :: (a -> Maybe b) -> [a] -> Maybe b
findFirstJust _ [] = Nothing
findFirstJust f (x:xs) =
    case f x of
      Just y -> Just y
      Nothing -> findFirstJust f xs

findFirstJustM :: Monad m => (a -> m (Maybe b)) -> [a] -> m (Maybe b)
findFirstJustM f l = firstJustM (map f l)

firstJustM :: Monad m => [m (Maybe a)] -> m (Maybe a)
firstJustM [] = return Nothing
firstJustM (m:ms) =
    do x <- m
       case x of
         Just _ -> return x
         Nothing -> firstJustM ms
    
concatMapM :: Monad m => (a -> m [b]) -> [a] -> m [b]
concatMapM f l =
    do ll <- mapM f l
       return (concat ll)

findAndDelete :: (a -> Bool) -> [a] -> Maybe (a, [a])
findAndDelete _ [] = Nothing
findAndDelete f (x:xs) = 
    if f x then Just (x, xs)
           else do (y,ys) <- findAndDelete f xs
                   return (y, x:ys)

disjointFrom :: Ord a => Set a -> Set a -> Bool
disjointFrom s1 s2 = Set.null (s1 `Set.intersection` s2)

count :: Eq a => a -> [a] -> Int
count _ [] = 0
count x (y:ys) | x == y = 1 + count x ys
               | otherwise = count x ys

mapMaybeM :: Monad m => (a -> m (Maybe b)) -> [a] -> m [b]
mapMaybeM _ [] = return []
mapMaybeM f (x:xs) =
    do my <- f x
       ys <- mapMaybeM f xs
       case my of
         Just y -> return (y:ys)
         Nothing -> return ys

mread :: (Read a, MonadPlus m) => String -> m a
mread s = 
    case reads s of 
      [] -> mzero 
      xs -> return . fst . head $ xs

map2 :: (a -> b -> c) -> [a] -> [b] -> [c]
map2 f (x:xs) (y:ys) = f x y : map2 f xs ys
map2 _ _ _ = []
