package javagi.runtime;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import junit.framework.TestCase;

public class CoreRTTest extends TestCase {

    private Integer intArg = new Integer(1);
    private Float floatArg = new Float(2.0);
    private ArrayList<String> arrayListArg = new ArrayList<String>();
    private LinkedList<String> linkedListArg = new LinkedList<String>();
    private String stringArg = "JavaGI";
    
    private Implementation impl = new Implementation(
            Comparable.class,
            new Class<?>[]{Number.class, List.class},
            null);
        
    public void testMatches() {
        int[] dispatchVector = new int[]{0,0, 1,3, 0,1, 1,4};
        assertTrue(CoreRT.matches(impl, dispatchVector,
                new Object[]{intArg, floatArg, null, arrayListArg, linkedListArg}));
        assertTrue(CoreRT.matches(impl, dispatchVector,
                new Object[]{intArg, floatArg, stringArg, arrayListArg, linkedListArg}));
        
        assertFalse(CoreRT.matches(impl,dispatchVector,
                new Object[]{stringArg, floatArg, null, arrayListArg, linkedListArg}));
        assertFalse(CoreRT.matches(impl,dispatchVector,
                new Object[]{intArg, stringArg, null, arrayListArg, linkedListArg}));
        assertFalse(CoreRT.matches(impl, dispatchVector,
                new Object[]{intArg, floatArg, null, stringArg, linkedListArg}));
        assertFalse(CoreRT.matches(impl, dispatchVector,
                new Object[]{intArg, floatArg, null, arrayListArg, stringArg}));
        
        try {
            CoreRT.matches(impl, dispatchVector,
                new Object[]{null, floatArg, stringArg, arrayListArg, linkedListArg});
            fail("NullPointerException expected");
        } catch (NullPointerException e) {
        }

        try {
            CoreRT.matches(impl, dispatchVector,
                new Object[]{intArg, null, stringArg, arrayListArg, linkedListArg});
            fail("NullPointerException expected");
        } catch (NullPointerException e) {
        }

        try {
            CoreRT.matches(impl, dispatchVector,
                new Object[]{intArg, floatArg, stringArg, null, linkedListArg});
            fail("NullPointerException expected");
        } catch (NullPointerException e) {
        }
        
        try {
            CoreRT.matches(impl, dispatchVector,
                new Object[]{intArg, floatArg, stringArg, arrayListArg, null});
            fail("NullPointerException expected");
        } catch (NullPointerException e) {
        }
    }

}
